# Example : interactive multistages 2G analysis

> [GWBurst: How to do an interactive multistages 2G analysis](https://gwburst.gitlab.io/documentation/latest/html/faq.html#how-to-do-an-interactive-multistages-2g-analysis)


## Setup

Copy the example files to the working directory

```bash
cp -r $CWB_SOURCE/tools/cwb/examples/ADV_SIM_NSNS_L1H1V1_MultiStages2G  MultiStages2G
cd MultiStages2G
pycwb_gen_config --cwb_install <path to cwb install> --cwb_source <path to cwb source> --work_dir <path to work dir>
ipython
```

> Optional: Replace the `exit(0);` in line 25 of file `macro/CreateFramesNOISE.c` to `return 0;` to prevent python exit
> Optional: Replace the `gSystem->Exit(0);` in line 92 of file `macro/cwb_inet2G.c` to `return 0;` to prevent python exit

Modify `config.ini` with your own setup

Initialize ROOT with class `pycWB` (_Recommanded_)

```python
import sys
sys.path.append('..') # add pycWB parent path
from pycwb import pycWB

cwb = pycWB('./config.ini')
ROOT = cwb.ROOT
gROOT = cwb.gROOT
configs = cwb.config
```

or with pure function

```python
import sys
sys.path.append('..') # add pycWB parent path
from pycwb import root_setup

ROOT, gROOT, gSystem, gStyle, configs = root_setup.init('./config.ini')
```

Run the setup to create frame file

```python
import  os, glob
from os.path import dirname, exists
from pathlib import Path

def setup(ROOT, gROOT):
	# setup plugin path
	if not exists('plugins'): os.symlink(f"{configs['CWB']['CWB_INSTALL']}/etc/cwb/plugins",'plugins')
	gROOT.LoadMacro("macro/CreateFramesNOISE.C")
	ROOT.CreateFramesNOISE()

	for dir in ['input', 'data', 'tmp/public_html/reports', 'tmp/condor', 'tmp/node', 'report/dump']:
		os.makedirs(dir, exist_ok=True)

	for det in ['H1','L1','V1']:
		with open(f"input/{det}.frames", 'w') as t:
			t.write(glob.glob(f"{os.getcwd()}/frames/*/{det}*.gwf")[0])

	with open('input/inspiral.in', 'w') as t:
		t.write("931158200    931158600\n")

setup(ROOT, gROOT)
```

## INIT STAGE

Read Config / CAT1-2 / User Plugin

```python
cwb.cwb_inet2G(1,'config/user_parameters.C', 'INIT')
```

Addtional method: with pure function

```python
from pycwb import cwb_interface
cwb_interface.cwb_inet2G(ROOT, gROOT, configs,1,'config/user_parameters.C', 'INIT')
```

### Check file

```python
ifile = ROOT.TFile.Open('data/init_931158208_192_MultiStages2G_job1.root')
ihistory = ifile.Get("history")
ihistory.DumpToTextFile('report/dump/init_931158208_192_MultiStages2G_job1.history')
```

## STRAIN STAGE

Read gw-strain / MDC data frames(or On The Fly MDC)

```python
cwb.cwb_inet2G(0,'data/init_931158208_192_MultiStages2G_job1.root', 'STRAIN')
```

produce L1 PSD and save the plot under the report/dump directory

```python
inet_option = '--tool psd --ifo L1 --type strain --draw true --save true'
cwb.cwb_inet2G(0,'data/init_931158208_192_MultiStages2G_job1.root', 'STRAIN', inet_option=inet_option)
```

display H1 noise with `FrDisplay <frdisplay.html#frdisplay>`__

```python
inet_option = '--tool frdisplay --hpf 50 --decimateby 8 --ifo H1 --type strain'
cwb.cwb_inet2G(0,'data/init_931158208_192_MultiStages2G_job1.root', 'STRAIN', inet_option=inet_option)
```

display injected waveforms in ROOT browser (Time/FFT/TF domain)

```python
inet_option = '--tool inj --draw true'
cwb.cwb_inet2G(0,'data/init_931158208_192_MultiStages2G_job1.root', 'STRAIN', inet_option=inet_option)
```

> right mouse button doesn't work for me

## CSTRAIN STAGE

Data Conditioning (Line Removal & Whitening)

```python
cwb.cwb_inet2G(0,'data/init_931158208_192_MultiStages2G_job1.root', 'CSTRAIN')
```

produce L1 T/F nRMS and visualize the plot in the ROOT browser ( )

```python
inet_option = '--tool nrms --ifo L1 --type strain --draw true'
cwb.cwb_inet2G(0,'data/strain_931158208_192_MultiStages2G_job1.root', 'CSTRAIN', inet_option=inet_option)
```

produce whitend H1 PSD and save the plot under the report/dump directory

```python
inet_option = '--tool psd --ifo H1 --type white --draw true --save true'
cwb.cwb_inet2G(0,'data/strain_931158208_192_MultiStages2G_job1.root', 'CSTRAIN', inet_option=inet_option)
```

produce whitend H1 TF WDM and visualize the plot in the ROOT browser

```python
inet_option = '--tool wdm --ifo L1 --type white --draw true'
cwb.cwb_inet2G(0,'data/strain_931158208_192_MultiStages2G_job1.root', 'CSTRAIN', inet_option=inet_option)
```


## COHERENCE STAGE

TF Pixels Selection

```python
cwb.cwb_inet2G(0,'data/cstrain_931158208_192_MultiStages2G_job1.root', 'COHERENCE')
```

produce TF WDM of maximum energy (before the pixel selection) at level 8 and visualize the plot in the ROOT browser

```python
inet_option = '--tool emax --level 8  --draw true'
cwb.cwb_inet2G(0,'data/cstrain_931158208_192_MultiStages2G_job1.root', 'COHERENCE', inet_option=inet_option)
```

## SUPERCLUSTER STAGE

Clustering & Cluster Selection

```python
cwb.cwb_inet2G(0,'data/coherence_931158208_192_MultiStages2G_job1.root', 'SUPERCLUSTER')
```

produce TF WDM of L1 sparse map and visualize the plot in the ROOT browser

```python
inet_option = '--tool sparse --type supercluster --ifo L1 --draw true'
cwb.cwb_inet2G(0,'data/coherence_931158208_192_MultiStages2G_job1.root', 'SUPERCLUSTER', inet_option=inet_option)
```

## LIKELIHOOD STAGE

Event Reconstruction & Output Parameters

```python
cwb.cwb_inet2G(0,'data/supercluster_931158208_192_MultiStages2G_job1.root', 'LIKELIHOOD')
```

produce TF WDM of L1 sparse map and visualize the plot in the ROOT browser

```python
# !rm data/wave_931158208_192_MultiStages2G_120_job1.root
inet_option = '--tool sparse --type likelihood --ifo L1 --draw true --save true'
cwb.cwb_inet2G(0,'data/supercluster_931158208_192_MultiStages2G_job1.root', 'LIKELIHOOD', inet_option=inet_option)
```


## Dump the event

```python
import os
os.environ['CWB_DUMP_EVT_FILE_NAME'] = 'data/wave_931158208_192_MultiStages2G_120_job1.root'
cwb.cwb_load_macro('cwb_dump_events.c')
```

